#include <stdlib.h>
#include "mpi.h"
#include "mpe.h"
#include "p_m_gen.h"

#define DEBUG 0

typedef struct {
  int x, y, width, height;
} to_do_item;

/* shading techniques: */
#define MAXITER_SHADE 1
/* (n==maxiter) */
#define EVEN_SHADE    2
/* (n==maxiter || !(n%2)) */

#define TO_DO_MSG  42
#define DONE_MSG   43
#define DATA_MSG   44
#define RETIRE_MSG 45

#define NUMCOLORS 32

/* macro for assigning to_do_list elements */
#define TDL_ASSIGN(a,b) { \
  to_do_list[n].x = x; \
  to_do_list[n].y = y; \
  to_do_list[n].width = (a); \
  to_do_list[n++].height = (b); \
}


#define IS_SET(z) (\
  (shadeType==MAXITER_SHADE) ? ((z)==maxiter) : \
  (shadeType==EVEN_SHADE)    ? ((z)==maxiter || !((z)%2)) : (z))



void ProcessArgs( argc, argv, width, height, rmin, rmax, imin, imax,
		  boundary, maxiter, bw, shadeType, chunksPer )
int argc, *width, *height, *maxiter, *bw, *shadeType, *chunksPer;
double *boundary;
NUM *rmin, *rmax, *imin, *imax;
char **argv;
{
  int argNum;
  double temp;

  for (argNum=1; argNum < argc; argNum++) {
#if DEBUG
fprintf( stderr, "argv[%d] = %s\n", argNum, argv[argNum] );
#endif
    if (!strcmp( argv[argNum], "-boundary" )) {
      sscanf( argv[++argNum], "%lf", boundary );
      
    } else if (!strcmp( argv[argNum], "-rmin" )) {
      sscanf( argv[++argNum], "%lf", &temp );
      NUM_PTR_ASSIGN( rmin, DOUBLE2NUM( temp ) );
	
    } else if (!strcmp( argv[argNum], "-rmax" )) {
      sscanf( argv[++argNum], "%lf", &temp );
      NUM_PTR_ASSIGN( rmax, DOUBLE2NUM( temp ) );
	
    } else if (!strcmp( argv[argNum], "-imin" )) {
      sscanf( argv[++argNum], "%lf", &temp );
      NUM_PTR_ASSIGN( imin, DOUBLE2NUM( temp ) );
	
    } else if (!strcmp( argv[argNum], "-imax" )) {
      sscanf( argv[++argNum], "%lf", &temp );
      NUM_PTR_ASSIGN( imax, DOUBLE2NUM( temp ) );
	
    } else if (!strcmp( argv[argNum], "-maxiter" )) {
      sscanf( argv[++argNum], "%d", maxiter );

    } else if (!strcmp( argv[argNum], "-width" )) {
      sscanf( argv[++argNum], "%d", width );

    } else if (!strcmp( argv[argNum], "-height" )) {
      sscanf( argv[++argNum], "%d", height );

    } else if (!strcmp( argv[argNum], "-chunks" )) {
      sscanf( argv[++argNum], "%d", chunksPer );

    } else if (!strcmp( argv[argNum], "-bw" )) {
      *bw=1;

    } else if (!strcmp( argv[argNum], "-shade" )) {
      if (!strcmp( argv[++argNum], "even" )) {
	*shadeType = EVEN_SHADE;
      } else if (!strcmp( argv[argNum], "maxiter" )) {
	*shadeType = MAXITER_SHADE;
      } else {
	printf( "Unknown shade type \"%s\".  Using \"maxiter\"\n",
	        argv[argNum] );
	*shadeType = MAXITER_SHADE;
      }

    } else if (!strcmp( argv[argNum], "-h" )) {
      printf( "Options recognized by %s:\n", argv[0] );
      printf( "(defaults are in parentheses () )\n" );
      printf( "\t-o        (<stdout>) output file\n" );
      printf( "\t-width    (500) width of computed area in points\n" );
      printf( "\t-height   (500) height of computed area in points\n" );
      printf( "\t-boundary (4.0) boundary point for M-set computation algorithm\n" );
      printf( "\t-maxiter  (100) maximum # of iterations for M-set compuptation algorithm\n" );
      printf( "\t-rmin     (-2.0) minimum real coordinate of computed area\n" );
      printf( "\t-rmax     (2.0) maximum real coordinate of computed area\n" );
      printf( "\t-imin     (-2.0) minimum imaginary coordinate of computed area\n" );
      printf( "\t-imax     (2.0) maximum imaginary coordinate of computed area\n" );
      printf( "\t-chunks   (3) estimated number chunks assigned to each process\n" );
      printf( "\t-bw       Draw in black and white\n" );
      printf( "\t-shade    (maxiter) what to fill when drawing in B&W-\n" );
      printf( "\t              maxiter - where 'iter' = maxiter\n" );
      printf( "\t              even - where 'iter' is even\n" );
      putchar( '\n' );
      exit( 0 );

    }
  }

fprintf( stderr, "From ProcessArgs: %lf %lf %lf %lf\n", *rmin, *rmax, *imin, *imax );
}



to_do_item *WhatToDo( np, width, height, chunksPer, nchunks, maxChunkLen )
int np, width, height, *nchunks, *maxChunkLen, chunksPer;

/* WhatToDo returns a 'todo' list of what chunks need to be calculated.
   It splits up the area to be calculated into at least np*CHUNKS_PER_PROC
   squares.
*/
{
  int chunkSide, smallerSide, n, x, y;
  to_do_item *to_do_list;

  np--;									     /* actually only have use of np-1 processes */
  smallerSide = (width<height) ? width : height;
  *nchunks = 1;
  for (n=1; *nchunks<np*chunksPer; n++) {
    chunkSide = (smallerSide-1)/n+1;
    *nchunks = ((width-1)/chunkSide+1) * ((height-1)/chunkSide+1);
  }

  to_do_list = (to_do_item *) malloc( sizeof( to_do_item ) * *nchunks );
  n=0;
  for (y=0; (y+chunkSide-1)<height; y+=chunkSide) {
    for (x=0; (x+chunkSide-1)<width; x+=chunkSide) {
      TDL_ASSIGN( chunkSide, chunkSide );
    }
    if ( width-x ) 
      TDL_ASSIGN( width-x, chunkSide );
  }
  if ( height-y ) {
    for (x=0; (x+chunkSide-1)<width; x+=chunkSide) {
      TDL_ASSIGN( chunkSide, height-y );
    }
    if ( width-x ) 
      TDL_ASSIGN( width-x, height-y );
  }

  *maxChunkLen = chunkSide * chunkSide;
  return to_do_list;
}



void OutputData( outf, iterationField, width, height, rmin, rmax, imin, imax,
	        maxiter, boundary )
FILE *outf;
int *iterationField, width, height, maxiter;
double boundary;
NUM rmin, rmax, imin, imax;
{
  int i, j;
  fprintf( outf, "%d %d %d %lf %lf %lf %lf %lf\n", width, height, maxiter,
	  boundary, NUM2DOUBLE( rmin ), NUM2DOUBLE( rmax ),
	  NUM2DOUBLE( imin ), NUM2DOUBLE( imax ) );
  for (j=0; j<height; j++) {
    for (i=0; i<width; i++) {
      fprintf( outf, "%3d ", *iterationField++ );
    }
    putc( '\n', outf );
  }
}


void OutputBlock( outf, procNum, field, width, height )
FILE *outf;
int procNum, *field, width, height;
{
  int a, b;
  fprintf( outf, "From proc %d:\n", procNum );
  for (b=0; b<height; b++) {
    for (a=0; a<width; a++) {
      fprintf( outf, "%3d ", *field++ );
    }
    putc( '\n', outf );
  }
  putc( '\n', outf );
}



void DrawChunk( graph, chunkField, x, y, width, height, maxiter, bw, shadeType )
MPE_XGraph *graph;
int *chunkField, x, y, width, height, bw, shadeType;
{
  int a, b, xlimit, ylimit;

#if DEBUG
MPI_Comm_rank( MPI_COMM_WORLD, &a );
fprintf( stderr, "[%d] DrawChunk (%d %d %d %d)\n", a, x, y, width, height );
#endif

  xlimit = x+width;
  ylimit = y+height;

  for (b=y; b<ylimit; b++) {
    for (a=x; a<xlimit; a++) {
      if (bw) {
	if (IS_SET(*chunkField)) {
	  MPE_DrawPoint( graph, a, b, 1 );
	}
      } else {
	MPE_DrawPoint( graph, a, b, (*chunkField==maxiter) ?
		       0 : (*chunkField % NUMCOLORS));
      }
      chunkField++;
    }
  }
  MPE_Update( graph );
}




void Master_Process(to_do_list, nchunks)
to_do_item *to_do_list;
int nchunks;
{
  int procNum, nbusy, nextToDo, *busyList, np, ndrawing;
  /* nextToDo - next chunk to calculate */
  /* procNum - casual use when a process number is used */
  /* busyList - kept by proc 0, list of what chunk each proc is working on */
  /* nbusy - number of busy processes */
  /* np - number of processes */


  MPI_Status status;

  MPI_Comm_size( MPI_COMM_WORLD, &np );
  busyList = (int *) malloc( sizeof( int ) * (np-1) );			     /* allocate memory for a list list that keeps track */
									     /* of what each process is doing */
  nextToDo = 0;
  procNum = 1;
  nbusy = 0;
  while (procNum<np) {
    MPI_Send( to_do_list + nextToDo, sizeof(to_do_item), MPI_BYTE,	     /* give each process a job */
	     procNum, TO_DO_MSG, MPI_COMM_WORLD );
#if DEBUG
fprintf( stderr, "Master telling %d to do (%d %d %d %d)\n", procNum,
	 to_do_list[nextToDo].x,
	 to_do_list[nextToDo].y,
	 to_do_list[nextToDo].width,
	 to_do_list[nextToDo].height);
#endif
    busyList[procNum++] = nextToDo++;
    nbusy++;								     /* keep track of how many processes are busy */
  }

  while (nbusy) {							     /* until everyone's done */
    MPI_Recv( &procNum, 1, MPI_INT, MPI_ANY_SOURCE, DONE_MSG,		     /* receive a message */
	     MPI_COMM_WORLD, &status );
#if DEBUG
fprintf( stderr, "Slave %d says it's done\n", procNum );
#endif

    if (nextToDo<nchunks) {						     /* if there's more work to do, */
      MPI_Send( to_do_list + nextToDo, sizeof(to_do_item), MPI_BYTE,	     /* tell the process that just finished its chunk */
	       procNum, TO_DO_MSG, MPI_COMM_WORLD );			     /* to start working on another */
#if DEBUG
fprintf( stderr, "Master assigns (%d %d %d %d) to %d\n", 
		     to_do_list[nextToDo].x,
		     to_do_list[nextToDo].y,
		     to_do_list[nextToDo].width,
		     to_do_list[nextToDo].height, procNum);
#endif

      busyList[procNum] = nextToDo++;					     /* mark one more job as in progress */
    } else {								     /* if no more work needs to be assigned... */

      /* might want a condition that re-assigns chunks that have been out for a long time, */
      /* in case that process is overloaded  */

      MPI_Send( to_do_list, sizeof( to_do_item ), MPI_BYTE, procNum,	     /* tell this process to go home */
	       RETIRE_MSG, MPI_COMM_WORLD );
#if DEBUG
fprintf( stderr, "Master gives process %d an honorable discharge\n", procNum );
#endif
      busyList[procNum] = -1;						     /* mark this process as idle */
      nbusy--;							     /* decrement the number of busy processes */
    }
  } /* while (nbusy) */

  while (1) sleep( 1000 );
}




void Slave_Process( graph, boundary, maxiter, width, height,
		    rmin, rmax, imin, imax,
		    maxChunkLen, bw, shadeType )
MPE_XGraph *graph;
int maxiter, maxChunkLen, bw, shadeType;
NUM rmin, rmax, imin, imax;
double boundary;

{
  int working=1, *chunkField, myid;
  /* working - 1=keep working, 0=all done */
  /* chunkField - data storage area */
  /* myid - this process' rank */

  to_do_item thisChunk;
  /* coordinates for the current chunk */

  MPI_Status status;
  /* status of MPI_Recv calls */

  FILE *outf;
  /* output file for debugging info */

  char outfilename[20];
  /* output filename */

  MPI_Comm_rank( MPI_COMM_WORLD, &myid );

#if DEBUG
sprintf( outfilename, "proc%d.out", myid );
outf = fopen( outfilename, "w" );
fprintf( stderr, "[%d] CALLED WITH: %lf %lf %lf %lf \n",
	myid, rmin, rmax, imin, imax );
#endif

  chunkField = (int *) malloc( sizeof( int ) * maxChunkLen );		     /* allocate memory for a buffer in which to keep data */

  Mbrot_Settings( boundary, maxiter );					     /* initialize the M-set calculation routines */

  while (working) {							     /* loop until retirement */
    MPI_Recv( &thisChunk, sizeof( to_do_item ), MPI_BYTE, 0, MPI_ANY_TAG,    /* get next task from master */
	     MPI_COMM_WORLD, &status );
    if (status.MPI_TAG==TO_DO_MSG) {					     /* if master says to calculate another chunk, */
#if DEBUG
fprintf( stderr, "[%d] slave driver... (%d %d %d %d)\n", myid, thisChunk.x,
	thisChunk.y, thisChunk.width, thisChunk.height );
#endif
      CalcField( outf, MbrotCalcIter, chunkField,			     /* calculate the chunk */
		rmin, rmax, imin, imax,
		0, width-1, 0, height-1,
		thisChunk.x, thisChunk.x + thisChunk.width-1,
		thisChunk.y, thisChunk.y + thisChunk.height-1 );

#if DEBUG
OutputBlock( outf, myid, chunkField, thisChunk.width, thisChunk.height );
#endif

      MPI_Send( &myid, 1, MPI_INT, 0, DONE_MSG, MPI_COMM_WORLD );	     /* tell the master "I'm done" */
      DrawChunk( graph, chunkField, thisChunk.x, thisChunk.y,		     /* draw this chunk */
		 thisChunk.width, thisChunk.height, maxiter,
		 bw, shadeType );

#if DEBUG
fprintf( stderr, "[%d] done with chunk\n", myid );
#endif

    } else if (status.MPI_TAG==RETIRE_MSG) {				     /* if the master says to retire, */

#if DEBUG
fprintf( stderr, "[%d] retiring\n", myid );
#endif

      working = 0;							     /* quit working */
    }
  } /* while (working ) */
}




int main( argc, argv )
int argc;
char **argv;
{
  MPI_Status status;
  /* status of MPI_Recv calls */

  MPE_XGraph *graph;
  /* handle to the X display */

  MPE_Color colorArray[NUMCOLORS];

  int myid, np, str_len, nchunks, maxChunkLen, ncolors, chunksPer=6;
  /* myid - this process' rank */
  /* np - number of processes */
  /* str_len - length of DISPLAY variable */
  /* nchunks - total number of chunks */
  /* maxChunkLen - maximum number of points in each chunk */
  /* ncolors - number of colors */
  /* chunksPer - minimum number of chunks for each slave process */

  char *display;
  /* string form of the DISPLAY variable */

  double boundary=4;
  /* boundary - boundary point in Mandelbrot set calculation sequence */

  NUM rmin, rmax, imin, imax;
  /* rmin, rmax, imin, imax - the area to be calculated */

  int maxiter=100, width=500, height=500, bw=0, shadeType=MAXITER_SHADE;
  /* maxiter - maximum # of iterations for mbrot calculation */
  /* width, height - dimensions of the whole field to be calculated */
  /* bw - 1 = B&W display, 0 = color */
  /* shadeType - EVEN_SHADE or MAXITER_SHADE, determines what is colored in B&W mode */

  to_do_item *to_do_list;
  /* array of chunk coordinates, to know what needs to be done */

  MPI_Init( &argc, &argv );

  MPI_Comm_rank( MPI_COMM_WORLD, &myid );
  MPI_Comm_size( MPI_COMM_WORLD, &np );

  NUM_ASSIGN( rmin, INT2NUM( -2 ) );
  NUM_ASSIGN( rmax, INT2NUM(  2 ) );
  NUM_ASSIGN( imin, INT2NUM( -2 ) );
  NUM_ASSIGN( imax, INT2NUM(  2 ) );

  if (!myid) {
    ProcessArgs( argc, argv, &width, &height, &rmin, &rmax, &imin, &imax,
		 &boundary, &maxiter, &bw, &shadeType, &chunksPer );
    display = getenv( "DISPLAY" );
    if (!display) {
      printf ("DISPLAY environment variable not set.  Exiting.\n\n");
      exit( -1 );
    }
    str_len = strlen( display ) + 1;
    MPI_Bcast( &str_len, 1, MPI_INT, 0, MPI_COMM_WORLD );
    MPI_Bcast( display, str_len, MPI_CHAR, 0, MPI_COMM_WORLD );
  } else {
    MPI_Bcast( &str_len, 1, MPI_INT, 0, MPI_COMM_WORLD );
    display = (char *) malloc( sizeof( char ) * str_len );
    MPI_Bcast( display, str_len, MPI_CHAR, 0, MPI_COMM_WORLD );
  }

  to_do_list = WhatToDo( np, width, height, chunksPer,			     /* split up region into to-do items, or 'chunks' */
			 &nchunks, &maxChunkLen );

  MPI_Bcast( &width,  1, MPI_INT, 0, MPI_COMM_WORLD );
  MPI_Bcast( &height, 1, MPI_INT, 0, MPI_COMM_WORLD );
  MPI_Bcast( &rmin, sizeof( NUM ), MPI_BYTE, 0, MPI_COMM_WORLD );
  MPI_Bcast( &rmax, sizeof( NUM ), MPI_BYTE, 0, MPI_COMM_WORLD );
  MPI_Bcast( &imin, sizeof( NUM ), MPI_BYTE, 0, MPI_COMM_WORLD );
  MPI_Bcast( &imax, sizeof( NUM ), MPI_BYTE, 0, MPI_COMM_WORLD );
  MPI_Bcast( &boundary,       1, MPI_DOUBLE, 0, MPI_COMM_WORLD );
  MPI_Bcast( &maxiter,        1, MPI_INT,    0, MPI_COMM_WORLD );
  MPI_Bcast( &maxChunkLen,    1, MPI_INT,    0, MPI_COMM_WORLD );
  MPI_Bcast( &bw,             1, MPI_INT,    0, MPI_COMM_WORLD );
  MPI_Bcast( &shadeType,      1, MPI_INT,    0, MPI_COMM_WORLD );

  MPE_OpenGraphics( &graph, MPI_COMM_WORLD, display, -1, -1, width, height, 0 );
  MPE_MakeColorArray( graph, NUMCOLORS, colorArray );
  MPE_Update( graph );

#if DEBUG
fprintf( stderr, "[%d] width %d, height %d, rmin %lf, rmax %lf, imin %lf, imax %lf, boundary %lf, maxiter %d\n",
	myid, width, height, rmin, rmax, imin, imax, boundary, maxiter );
#endif

  if (!myid) {
    Master_Process( to_do_list, nchunks );
  } else {
    Slave_Process( graph, boundary, maxiter, width, height,
		   rmin, rmax, imin, imax, maxChunkLen, bw, shadeType );
  }
  MPI_Finalize();
}
