/*
 *  $Id: mpir.h,v 1.17 1994/04/15 23:17:01 gropp Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

/* include file for the MPIR implementation of MPI */

#include <stdio.h>

#ifndef _MPIR_INCLUDE
#define _MPIR_INCLUDE

#include "mpid.h"

/*****************************************************************************
*                           MPIR DATA STRUCTURES                             *
*****************************************************************************/

typedef struct _MPIR_DATATYPE {
    MPIR_NODETYPE dte_type; /* type of datatype element this is */
    MPIR_BOOL    committed; /* whether committed or not */
    int          is_contig; /* whether entirely contiguous */
    int              basic; /* Is this a basic type */
    int             ub, lb; /* upper/lower bound of type */
    int             extent; /* extent of this datatype */
    int               size; /* size of type */
    int           elements; /* number of basic elements */
    int          ref_count; /* nodes depending on this node */
	int              align; /* alignment needed for start of datatype */
    int              count; /* replication count */
    MPI_Aint        stride; /* stride, for VECTOR and HVECTOR types */
    int           blocklen; /* blocklen, for VECTOR and HVECTOR types */
    int           *indices; /* array of indices, for (H)INDEXED */
    int         *blocklens; /* array of blocklens for (H)INDEXED */
    struct _MPIR_DATATYPE *old_type;  /* type this type is built of, if 1 */
    struct _MPIR_DATATYPE **old_types;/* array of types, for STRUCT */
} MPIR_DATATYPE;

typedef enum {
    MPIR_SEND,
    MPIR_RECV
} MPIR_OPTYPE;

/* MPIR_COMMON is a subset of the handle structures that contain common
   elements.  MAKE SURE THAT ANY CHANGES IN THE HANDLES PRESERVES
   THIS ORDER! */
typedef struct {
    MPIR_OPTYPE handle_type;
    int         partner;        /* source or destination, depending on type */
    int         tag;
    int         contextid;
    MPIR_BOOL   completed;
    MPIR_DATATYPE *datatype;

    void *bufadd;       /* address of buffer */
    int buflen;         /* length of buffer at bufadd */
    int count;          /* requested count */

    char *bufpos;       /* position of next byte to be transferred */
    int  transfer_count;    /* count of bytes transferred to/from device */
    int  totallen;      /* length in local bytes of message */

    } MPIR_COMMON;

typedef struct {
    MPIR_OPTYPE handle_type;    /* send or receive */
    int         dest;       /* destination process for message */
    int         tag;        /* tag */
    int         contextid;  /* context id */
    MPIR_BOOL   completed;  /* whether operation is completed or not */
    MPIR_DATATYPE *datatype;    /* basic or derived datatype */
    void *bufadd;       /* address of buffer */
    int buflen;         /* length of buffer at bufadd */
    int count;          /* requested count */
    char *bufpos;       /* position of next byte to be transferred */
    int  transfer_count;    /* count of bytes transferred to/from device */
    int  totallen;      /* length in local bytes of message */

    MPIR_Mode mode;     /* mode: STANDARD, READY, SYNC */
    int  lrank;                 /* Rank in sending group */
    MPID_SHANDLE dev_shandle;   /* device's version of send handle */
} MPIR_SHANDLE;

typedef struct {
    MPIR_OPTYPE handle_type;    /* send or receive */
    int         source;     /* source process message */
    int     tag;        /* tag */
    int         contextid;  /* context id */
    MPIR_BOOL   completed;  /* Whether this receive has been completed */
    MPIR_DATATYPE *datatype;    /* basic or derived datatype */
    void *bufadd;       /* address of buffer */
    int buflen;         /* length of buffer at bufadd */
    int count;          /* requested count */
    char *bufpos;       /* position of next byte to be transferred */
    int  transfer_count;    /* count of bytes transferred to/from device */
    int  totallen;          /* length in local bytes of message */

    void *p;            /* Pointer to unexpected data */
    int  len;           /* Length of this data ? */
    int  actcount;      /* number of items actually read */
    MPID_RHANDLE dev_rhandle;   /* device's version of recv handle */
} MPIR_RHANDLE;

#define MPIR_HANDLES_DEFINED

typedef union {
    MPIR_OPTYPE type;       /* send or receive */
    MPIR_COMMON  chandle;       /* common fields */
    MPIR_SHANDLE shandle;
    MPIR_RHANDLE rhandle;
} MPIR_HANDLE;

typedef enum {
    MPIR_QSHANDLE,
    MPIR_QRHANDLE
} MPIR_QEL_TYPE;

typedef struct _MPIR_QEL {  /* queue elements */
    void *ptr;          /* queues can contain anything  */
    MPIR_QEL_TYPE qel_type; /* but we may need to know what */
    struct _MPIR_QEL *prev; /* previous queue element */
    struct _MPIR_QEL *next; /* next queue element */
} MPIR_QEL;

typedef struct {        /* header for queues of things like handles */
    MPIR_QEL *first;        /* first queue element */
    MPIR_QEL *last;     /* last queue element */
    int      currlen;       /* current length */
    int      maxlen;        /* maximum since beginning of run */
} MPIR_QHDR;

typedef struct _MPIR_FDTEL {    /* flat datatype elements */
    MPIR_NODETYPE      type;    /* one of the elementary data types */
    int                disp;    /* displacement */
    struct _MPIR_FDTEL *next;   /* pointer to next element */
} MPIR_FDTEL;

#define MPIR_MALLOC(size)            malloc(size)
#define MPIR_dump_datatype(datatype) MPIR_dump_dte(datatype,0)

/*****************************************************************************
*                                GLOBAL DATA                                 *
*****************************************************************************/

/* memory management for fixed-size blocks */
extern void *MPIR_shandles;     /* sbcnst MPIR_SHANDLES */
extern void *MPIR_rhandles;     /* sbcnst MPIR_RHANDLES */
extern void *MPIR_dtes;   /* sbcnst datatype elements */
extern void *MPIR_qels;   /* sbcnst queue elements */
extern void *MPIR_fdtels; /* sbcnst flat datatype elements */
extern void *MPIR_hbts;   /* sbcnst height balanced tree roots for cacheing */
extern void *MPIR_hbt_els;/* sbcnst height balanced tree nodes for cacheing */

/* pre-assigned datatypes */
extern MPIR_DATATYPE *MPIR_int_dte;
extern MPIR_DATATYPE *MPIR_float_dte;
extern MPIR_DATATYPE *MPIR_double_dte;
extern MPIR_DATATYPE *MPIR_long_dte;
extern MPIR_DATATYPE *MPIR_complex_dte; 
extern MPIR_DATATYPE *MPIR_dcomplex_dte; 
extern MPIR_DATATYPE *MPIR_short_dte;
extern MPIR_DATATYPE *MPIR_char_dte;
extern MPIR_DATATYPE *MPIR_byte_dte;
extern MPIR_DATATYPE *MPIR_uchar_dte;
extern MPIR_DATATYPE *MPIR_ushort_dte;
extern MPIR_DATATYPE *MPIR_ulong_dte;
extern MPIR_DATATYPE *MPIR_uint_dte;
extern MPIR_DATATYPE *MPIR_longdouble_dte;
extern MPIR_DATATYPE *MPIR_packed_dte;
extern MPIR_DATATYPE *MPIR_ub_dte;
extern MPIR_DATATYPE *MPIR_lb_dte;

/* queues */
extern MPIR_QHDR MPIR_posted_recvs;
extern MPIR_QHDR MPIR_unexpected_recvs;

/* MPIR routines */
extern MPIR_GROUP *MPIR_CreateGroup();

/* MPIR process id (from device) */
extern int MPIR_tid;

/* Fortran logical values */
extern int MPIR_F_TRUE, MPIR_F_FALSE;

#endif
