/*
 *  $Id: red_scat.c,v 1.14 1994/04/13 21:57:08 doss Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

#ifndef lint
static char vcid[] = "$Id: red_scat.c,v 1.14 1994/04/13 21:57:08 doss Exp $";
#endif /* lint */

#include <mpi.h>
#include "mpisys.h"
#include "coll.h"

/*@

MPI_Reduce_scatter - Combine values and scatter the results

Input Parameters:
. sendbuf - starting address of send buffer (choice) 
. recvcounts - integer array specifying the 
number of elements in result distributed to each process.
Array must be identical on all calling processes. 
. datatype - data type of elements of input buffer (handle) 
. op - operation (handle) 
. comm - communicator (handle) 

Output Parameter:
. recvbuf - starting address of receive buffer (choice) 
@*/
int MPI_Reduce_scatter ( sendbuf, recvbuf, recvcnts, datatype, op, comm )
void             *sendbuf;
void             *recvbuf;
int              *recvcnts;
MPI_Datatype      datatype;
MPI_Op            op;
MPI_Comm          comm;
{
  int   rank, size, i, count=0;
  MPI_Aint extent;
  int  *displs;
  void *buffer;
  int   errno = MPI_SUCCESS;
  int   flag;

  /* Check for invalid arguments */
  if ( ( (comm == MPI_COMM_NULL) && (errno = MPI_ERR_COMM) )   ||
       ( (op == MPI_OP_NULL)     && (errno = MPI_ERR_OP) )     ||
	   ( (recvbuf == sendbuf)    && (errno = MPI_ERR_BUFFER) ) )
    return MPIR_ERROR(comm, errno, "Error in MPI_REDUCE_SCATTER" );

  /* Check for intra-communicator */
  MPI_Comm_test_inter ( comm, &flag );
  if (flag) 
    return MPIR_ERROR(comm, MPI_ERR_COMM,
					  "Inter-communicator invalid in MPI_REDUCE_SCATTER");

  /* Determine the "count" of items to reduce and set the displacements*/
  MPI_Type_extent (datatype, &extent);
  MPI_Comm_size   (comm, &size);
  MPI_Comm_rank   (comm, &rank);

  /* Allocate the displacements and initialize them */
  displs = (int *)MALLOC(size*sizeof(int));
  if (!displs) 
      return MPIR_ERROR( comm, MPI_ERR_EXHAUSTED, 
			 "Out of space in MPI_REDUCE_SCATTER" );
  for (i=0;i<size;i++) {
    displs[i] = count;
    count += recvcnts[i];
  }

  /* Allocate a temporary buffer */
  buffer = (void *)MALLOC(extent*count);
  if (!buffer) 
      return MPIR_ERROR( comm, MPI_ERR_EXHAUSTED, 
			 "Out of space in MPI_REDUCE_SCATTER" );

  /* Reduce to 0, then scatter */
  MPI_Reduce   ( sendbuf, buffer, count, datatype, op, 0, comm);
  MPI_Scatterv ( buffer, displs, recvcnts, datatype, recvbuf, 
				 recvcnts[rank], datatype, 0, comm );
  
  /* Free the temporary buffers */
  FREE(buffer); FREE(displs);
  return (errno);
}
