/*
 *  $Id: scatterv.c,v 1.14 1994/04/13 21:57:08 doss Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

#ifndef lint
static char vcid[] = "$Id: scatterv.c,v 1.14 1994/04/13 21:57:08 doss Exp $";
#endif /* lint */

#include <mpi.h>
#include "coll.h"

/*@

MPI_Scatterv - Scatter a buffer in parts to all tasks in a group

Input Parameters:
. sendbuf - address of send buffer (choice, significant only at root) 
. sendcounts - integer array (of length group size) 
specifying the number of elements to send to each processor  
. displs - integer array (of length group size). Entry 
 i  specifies the displacement (relative to sendbuf  from
which to take the outgoing data to process  i  
. sendtype - data type of send buffer elements (handle) 
. recvcount - number of elements in receive buffer (integer) 
. recvtype - data type of receive buffer elements (handle) 
. root - rank of sending process (integer) 
. comm - communicator (handle) 

Output Parameter:
. recvbuf - address of receive buffer (choice) 

@*/
int MPI_Scatterv ( sendbuf, sendcnts, displs, sendtype, 
                   recvbuf, recvcnt,  recvtype, 
                   root, comm )
void             *sendbuf;
int              *sendcnts;
int              *displs;
MPI_Datatype      sendtype;
void             *recvbuf;
int               recvcnt;
MPI_Datatype      recvtype;
int               root;
MPI_Comm          comm;
{
  MPI_Status status;
  int        rank, size;
  int        errno = MPI_SUCCESS;
  int        flag;

  /* Get size and rank */
  MPI_Comm_size ( comm, &size );
  MPI_Comm_rank ( comm, &rank );

  /* Check for invalid arguments */
  if ( ( (comm            == MPI_COMM_NULL)  && (errno = MPI_ERR_COMM) )  ||
       ( (root            <  0)              && (errno = MPI_ERR_ROOT) )  || 
       ( (root            >= size)           && (errno = MPI_ERR_ROOT) )  ||
       ( ((recvcnt>0)&&(recvbuf==(void *)0)) && (errno = MPI_ERR_BUFFER) ) )
    return MPIR_ERROR( comm, errno, "Error in MPI_SCATTERV" );
  
  /* Check for intra-communicator */
  MPI_Comm_test_inter ( comm, &flag );
  if (flag) 
    return MPIR_ERROR(comm, MPI_ERR_COMM,
					  "Inter-communicator invalid in MPI_SCATTERV");
				  
  /* Switch communicators to the hidden collective */
  comm = comm->comm_coll;

  /* If I'm the root, then scatter */
  if ( rank == root ) {
    MPI_Aint extent;
    int      i;

    MPI_Type_extent(sendtype, &extent);
    for ( i=0; i<size; i++ )
      MPI_Send( (void *)((int)sendbuf+displs[i]*extent), sendcnts[i], 
        sendtype, i, MPIR_SCATTERV_TAG, comm);
  }

  /* Everybody receives from the root */
  MPI_Recv(recvbuf, recvcnt, recvtype, root, MPIR_SCATTERV_TAG, comm, &status);

  return (errno);
}


