/*
 *  $Id: dmpi.c,v 1.11 1994/04/13 21:44:30 gropp Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

#ifndef lint
static char vcid[] = "$Id: dmpi.c,v 1.11 1994/04/13 21:44:30 gropp Exp $";
#endif /* lint */

/*  dmpi.c - routines in mpir that are called by the device */

#include "mpi.h"
#include "mpisys.h"

#define MPIR_MIN(a,b) (a) < (b) ? (a) : (b)
#define DEVICE_PREFERS_MEMCPY 1

/* called by device when a message arrives.  Returns 1 if there is posted
 * receive, 0 otherwise.
 *
 * This puts the responsibility for searching the unexpected queue and
 * posted-receive queues on the device.  If it is operating asynchronously
 * with the user code, the device must provide the necessary locking mechanism.
 */

DMPI_msg_arrived( src, tag, context_id, dmpi_recv_handle, foundflag )
int               src, tag, *foundflag;
MPIR_CONTEXT      context_id;
MPIR_RHANDLE      **dmpi_recv_handle;
{
    int          found;
    MPIR_RHANDLE *handleptr;

    MPIR_search_posted_queue( src, tag, context_id, &found, 1, 
			      dmpi_recv_handle);
    if ( found )
    {
	*foundflag = 1;	
	/* note this overwrites any wild-card values in the posted handle */
	handleptr         	= *dmpi_recv_handle;
	handleptr->source 	= src;
	handleptr->tag  	= tag;
	/* count is set in the put and get routines */
    }
    else
    {
	/* allocate handle and put in unexpected queue */
	*dmpi_recv_handle       = ( MPIR_RHANDLE * ) SBalloc ( MPIR_rhandles );
	handleptr         	= *dmpi_recv_handle;
	handleptr->handle_type  = MPIR_RECV;
	handleptr->source 	= src;
	handleptr->tag  	= tag;
	handleptr->contextid    = context_id;
	handleptr->datatype     = MPI_BYTE;
	handleptr->completed    = NO;
	MPID_alloc_recv_handle( &(handleptr->dev_rhandle) );
	
	MPIR_enqueue( &MPIR_unexpected_recvs, (void * ) *dmpi_recv_handle,
                      MPIR_QSHANDLE );
	*foundflag = 0;
    }
}



/*
 *  This code implements one of the interfaces between the device and the
 *  MPI code for moving data between the two systems.  This code delivers
 *  to the device contiguous chuncks of data, packing or unpacking as
 *  required.
 */

/* get length of data to be sent
 */
DMPI_get_totallen( handle, len )
MPIR_SHANDLE *handle;
int          *len;
{
    *len = (handle->count)*(handle->datatype->extent);
				/* correct for contig messages, at least*/
    handle->totallen       = *len;
    handle->bufpos         = handle->bufadd;
    handle->transfer_count = 0;
}

/* tell mpi how many bytes have been read 
 */
DMPI_put_totallen( handle, len )
MPIR_RHANDLE *handle;
int          len;
{
    handle->totallen       = len;
    handle->bufpos         = handle->bufadd;
    handle->transfer_count = 0;
}

/* SENDS */
/* device has send buffer ready, wants mpi to copy data to it
 */
DMPI_get_into_contig( handle, addr, maxlen, actlen )
MPIR_SHANDLE *handle;
void         *addr;
int          maxlen;
int          *actlen;
{
int len_left = (handle->count)*(handle->datatype->extent) - 
               handle->transfer_count;

*actlen = MPIR_MIN(len_left,maxlen);
#ifdef DEVICE_PREFERS_MEMCPY
    memcpy( addr, handle->bufpos, *actlen );
#else
    /* device-defined macro for data movement, for example load-store through
       vector registers */
#endif
    handle->bufpos         += *actlen;
    handle->transfer_count += *actlen;
}

/* device will copy mpi's contiguous buffer, wants to know where it is
 */
DMPI_get_from_contig( handle, addr, maxlen, actlen )
MPIR_SHANDLE *handle;
void         **addr;
int          maxlen;
int          *actlen;
{
    
    *addr   = handle->bufpos;
    *actlen = (handle->count)*(handle->datatype->extent);
    if (*actlen > maxlen) *actlen = maxlen;
    handle->bufpos         += *actlen;
    handle->transfer_count += *actlen;
}

/* RECEIVES */
/* device will put data into mpi's contiguous buffer, wants to know where it is
 */
DMPI_put_into_contig( handle, addr, maxlen, actlen )
MPIR_RHANDLE *handle;
void         **addr;
int          maxlen;
int          *actlen;
{
    *addr                  = handle->bufpos;
    *actlen                = (handle->actcount)*(handle->datatype->extent);
    if (*actlen > maxlen) *actlen = maxlen;
    handle->bufpos         += *actlen;
    handle->transfer_count += *actlen;
}

/* device has contiguous buffer ready, wants mpi to copy data from it
 */
DMPI_put_from_contig( handle, addr, maxlen )
MPIR_RHANDLE *handle;
void         *addr;
int          maxlen;
{
    int actlen;

    actlen = MPIR_MIN( maxlen, handle->totallen - handle->transfer_count);

#ifdef DEVICE_PREFERS_MEMCPY
    memcpy( handle->bufpos, addr, actlen );
#else
    /* device-defined macro for data movement, for example load-store through
       vector registers */
#endif
    handle->bufpos         += actlen;
    handle->transfer_count += actlen;
    handle->actcount        =  handle->transfer_count / 
	(handle->datatype->extent);
}

/*
   Let the device tell the API that a handle can be freed (this handle
   was generated by an unexpected receive and inserted by DMPI_msg_arrived.
 */
DMPI_free_unexpected( dmpi_recv_handle  )
MPIR_RHANDLE      *dmpi_recv_handle;
{
SBfree( MPIR_rhandles, dmpi_recv_handle );
}
