/*
 *  $Id: datatypes.c,v 1.11 1994/04/15 23:15:37 gropp Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

#include "mpi.h"
#include "mpisys.h"

#ifndef MPIR_TRUE
#define MPIR_TRUE  1
#define MPIR_FALSE 0
#endif

/*@
    MPI_Type_extent - Return the extent of a datatype
@*/
int MPI_Type_extent( datatype, extent )
MPI_Datatype  datatype;
MPI_Aint     *extent;
{
  /* Check for bad datatype */
  if (datatype == MPI_DATATYPE_NULL) 
	return MPIR_ERROR( MPI_COMM_NULL, MPI_ERR_TYPE,
					  "Error in MPI_TYPE_EXTENT" );

  /* Assign the extent and return */
  (*extent) = datatype->extent;
  return (MPI_SUCCESS);
}


/*@
    MPI_Type_size - Return the number of bytes occupied by entries
                    in the datatype
@*/
int MPI_Type_size ( datatype, size )
MPI_Datatype  datatype;
MPI_Aint     *size;
{
  /* Check for bad datatype */
  if (datatype == MPI_DATATYPE_NULL) 
	return MPIR_ERROR( MPI_COMM_NULL, MPI_ERR_TYPE,
					  "Error in MPI_TYPE_SIZE" );

  /* Assign the size and return */
  (*size) = datatype->size;
  return (MPI_SUCCESS);
}


/*@
    MPI_Type_count - Return the number of "top-level" entries in 
	                 the datatype
@*/
int MPI_Type_count ( datatype, count )
MPI_Datatype  datatype;
int          *count;
{
  /* Check for bad datatype */
  if (datatype == MPI_DATATYPE_NULL) 
	return MPIR_ERROR( MPI_COMM_NULL, MPI_ERR_TYPE,
					  "Error in MPI_TYPE_COUNT" );

  /* Assign the count and return */
  (*count) = datatype->count;
  return (MPI_SUCCESS);
}


/*@
    MPI_Type_lb - Return the number of "top-level" entries in 
                  the datatype
@*/
int MPI_Type_lb ( datatype, lb )
MPI_Datatype  datatype;
int          *lb;
{
  /* Check for bad datatype */
  if (datatype == MPI_DATATYPE_NULL) 
	return MPIR_ERROR( MPI_COMM_NULL, MPI_ERR_TYPE,
					  "Error in MPI_TYPE_LB" );

  /* Assign the lb and return */
  (*lb) = datatype->lb;
  return (MPI_SUCCESS);
}


/*@
    MPI_Type_ub - Return the number of "top-level" entries in 
                  the datatype
@*/
int MPI_Type_ub ( datatype, ub )
MPI_Datatype  datatype;
int          *ub;
{
  /* Check for bad datatype */
  if (datatype == MPI_DATATYPE_NULL) 
	return MPIR_ERROR( MPI_COMM_NULL, MPI_ERR_TYPE,
					  "Error in MPI_TYPE_UB" );

  /* Assign the ub and return */
  (*ub) = datatype->ub;
  return (MPI_SUCCESS);
}

/*@
    MPI_Type_commit - Commits the datatype
@*/
int MPI_Type_commit ( datatype )
MPI_Datatype *datatype;
{
  /* This currently does nothing */
}

/*@
    MPI_Type_free - Frees the datatype
@*/
int MPI_Type_free ( datatype )
MPI_Datatype *datatype;
{
  int errno = MPI_SUCCESS;

  /* Check for bad arguments */
  if ( ( (datatype    == (MPI_Datatype *)0) && (errno = MPI_ERR_ARG) ) ||
	   ( ((*datatype) == MPI_DATATYPE_NULL) && (errno = MPI_ERR_TYPE) ) )
	return MPIR_ERROR( MPI_COMM_NULL, errno,
					  "Error in MPI_TYPE_FREE" );

  /* Don't free basic types */
  if ((*datatype)->basic == MPIR_TRUE)
	return (MPI_SUCCESS);

  /* Free datatype */
  if ( (*datatype)->ref_count <= 1 ) {

	/* Free malloc'd memory for various datatypes */
	if ( ((*datatype)->dte_type == MPIR_INDEXED)  ||
		 ((*datatype)->dte_type == MPIR_HINDEXED) || 
		 ((*datatype)->dte_type == MPIR_STRUCT)   ) {
	  FREE ( (*datatype)->indices );
	  FREE ( (*datatype)->blocklens );
	}

	/* Free the old_type if not a struct */
	if ((*datatype)->dte_type != MPIR_STRUCT)
	  MPI_Type_free ( &((*datatype)->old_type) );
	
	/* Free the old_types of a struct */
	if ((*datatype)->dte_type == MPIR_STRUCT) {
	  int i;

	  /* Decrease the reference count */
	  for (i=0; i<(*datatype)->count; i++)
		MPI_Type_free ( &((*datatype)->old_types[i]) );

	  /* Free the malloc'd memory */
	  FREE ( (*datatype)->old_types );
	}

	/* Free the datatype structure */
	SBfree ( MPIR_dtes, (*datatype) );
  }
  else 
	(*datatype)->ref_count--;

  (*datatype) = MPI_DATATYPE_NULL;
  return (errno);
}


/*+
    MPIR_Type_dup - Utility function used to "touch" a type
+*/
MPI_Datatype MPIR_Type_dup ( datatype )
MPI_Datatype datatype;
{
  datatype->ref_count++;
  return (datatype);
}


/*@
  MPI_Get_elements - Returns the number of basic elements
                     in a datatype
@*/
int MPI_Get_elements ( status, datatype, elements )
MPI_Status    status;
MPI_Datatype  datatype;
int          *elements;
{
  int count;

  /* Check for bad datatype */
  if (datatype == MPI_DATATYPE_NULL) 
	return MPIR_ERROR( MPI_COMM_NULL, MPI_ERR_TYPE,
					  "Error in MPI_GET_ELEMENTS" );
  
  /* Find the number of elements */
  MPI_Get_count (status, datatype, &count);
  if (count == MPI_UNDEFINED) {
	/* HACK ALERT -- the code in this if is not correct */
	/*               but for now ... */
	double cnt = 
	  (double) status.count_in_bytes / (double) datatype->size;
	(*elements) = (int) ( cnt * (double) datatype->elements );
  }
  else
	(*elements) = count * datatype->elements;

  return (MPI_SUCCESS);
}


/*@
    MPI_Type_contiguous - Create a contiguous datatype
@*/
int MPI_Type_contiguous( count, old_type, newtype )
int          count;
MPI_Datatype old_type;
MPI_Datatype *newtype;
{
  MPIR_DATATYPE  *dteptr;
  int errno = MPI_SUCCESS;

  /* Check for bad arguments */
  if ( ( (old_type == MPI_DATATYPE_NULL) && (errno = MPI_ERR_TYPE) )  ||
	   ( (old_type == (*newtype))        && (errno = MPI_ERR_TYPE) )  ||
	   ( (count   <= 0)                  && (errno = MPI_ERR_COUNT) ) )
	return MPIR_ERROR( MPI_COMM_NULL, errno,
					  "Error in MPI_TYPE_CONTIGUOUS" );
	
  /* Create and fill in the datatype */
  dteptr = (*newtype) = (MPI_Datatype) SBalloc( MPIR_dtes );
  dteptr->dte_type    = MPIR_CONTIG;
  dteptr->committed   = MPIR_FALSE;
  dteptr->basic       = MPIR_FALSE;
  dteptr->ref_count   = 1;
  dteptr->align       = old_type->align;
  dteptr->stride      = 1;
  dteptr->blocklen    = 1;
  dteptr->is_contig   = old_type->is_contig;
  dteptr->elements    = count * old_type->elements;

  /* Take care of contiguous vs non-contiguous case */
  if (old_type->is_contig) {
	dteptr->old_type  = MPIR_Type_dup (old_type->old_type);
	dteptr->count     = count * old_type->count;
  }
  else {
	dteptr->old_type  = MPIR_Type_dup (old_type);
	dteptr->count     = count;
  }

  /* Set the upper/lower bounds and the extent and size */
  dteptr->lb          = dteptr->old_type->lb;
  dteptr->extent      = dteptr->count * dteptr->old_type->extent;
  dteptr->ub          = dteptr->lb + dteptr->extent;
  dteptr->size        = dteptr->count * dteptr->old_type->size;
  
  return (errno);
}

/*@
    MPI_Type_vector - Create a vector (strided) datatype
@*/
int MPI_Type_vector( count, blocklen, stride, old_type, newtype )
int          count;
int 	     blocklen;
int 	     stride;
MPI_Datatype old_type;
MPI_Datatype *newtype;
{
  MPIR_DATATYPE  *dteptr;
  int errno = MPI_SUCCESS;

  /* Check for bad arguments */
  if ( ( (old_type == MPI_DATATYPE_NULL) && (errno = MPI_ERR_TYPE) )  ||
	   ( (old_type == (*newtype))        && (errno = MPI_ERR_TYPE) )  ||
	   ( (count   <= 0)                  && (errno = MPI_ERR_COUNT) ) ||
	   ( (blocklen <= 0)                 && (errno = MPI_ERR_ARG) )   )
	return MPIR_ERROR( MPI_COMM_NULL, errno,
					  "Error in MPI_TYPE_VECTOR" );
	
  /* Handle the case where blocklen & stride make a contiguous type */
  if ( (blocklen == stride) ||
	   (count    == 1) )
	return MPI_Type_contiguous ( count * blocklen, old_type, newtype );

  /* Reduce this to the hvector case */
  return MPI_Type_hvector ( count, blocklen, stride * old_type->extent,
						    old_type, newtype );
}

/*@
    MPI_Type_hvector - Create a vector (strided) datatype with offset in bytes
@*/
int MPI_Type_hvector( count, blocklen, stride, old_type, newtype )
int          count;
int 	     blocklen;
MPI_Aint     stride;
MPI_Datatype old_type;
MPI_Datatype *newtype;
{
  MPIR_DATATYPE  *dteptr;
  int errno = MPI_SUCCESS;

  /* Check for bad arguments */
  if ( ( (old_type == MPI_DATATYPE_NULL) && (errno = MPI_ERR_TYPE) )  ||
	   ( (old_type == (*newtype))        && (errno = MPI_ERR_TYPE) )  ||
	   ( (count   <= 0)                  && (errno = MPI_ERR_COUNT) ) ||
	   ( (blocklen <= 0)                 && (errno = MPI_ERR_ARG) )   )
	return MPIR_ERROR( MPI_COMM_NULL, errno,
					  "Error in MPI_TYPE_HVECTOR" );
	
  /* Handle the case where blocklen & stride make a contiguous type */
  if ( ((blocklen * old_type->extent) == stride) ||
	   (count                         == 1) )
	return MPI_Type_contiguous ( count * blocklen, old_type, newtype );

  /* Create and fill in the datatype */
  dteptr = (*newtype) = (MPI_Datatype) SBalloc( MPIR_dtes );
  dteptr->dte_type    = MPIR_HVECTOR;
  dteptr->committed   = MPIR_FALSE;
  dteptr->basic       = MPIR_FALSE;
  dteptr->is_contig   = MPIR_FALSE;
  dteptr->ref_count   = 1;
  dteptr->align       = old_type->align;
  dteptr->elements    = count * blocklen * old_type->elements;
  dteptr->stride      = stride;
  dteptr->blocklen    = blocklen;
  dteptr->old_type    = MPIR_Type_dup (old_type);
  dteptr->count       = count;

  /* Set the upper/lower bounds and the extent and size */
  dteptr->lb          = old_type->lb;
  dteptr->extent      = ((count-1) * stride) + (blocklen * old_type->extent);
  dteptr->ub          = dteptr->lb + dteptr->extent;
  dteptr->size        = count * blocklen * dteptr->old_type->size;
  
  return (errno);
}

/*@
    MPI_Type_indexed - Create an indexed datatype
@*/
int MPI_Type_indexed( count, blocklens, indices, old_type, newtype )
int           count;
int 	      blocklens[];
int 	      indices[];
MPI_Datatype  old_type;
MPI_Datatype *newtype;
{
  MPIR_DATATYPE  *dteptr;
  MPI_Aint ub, high, low;
  int      i, errno = MPI_SUCCESS;

  /* Check for bad arguments */
  if ( ( (old_type == MPI_DATATYPE_NULL) && (errno = MPI_ERR_TYPE) )  ||
	   ( (old_type == (*newtype))        && (errno = MPI_ERR_TYPE) )  ||
	   ( (count    <= 0)                 && (errno = MPI_ERR_COUNT) ) )
	return MPIR_ERROR( MPI_COMM_NULL, errno,
					  "Error in MPI_TYPE_INDEXED" );
	
  /* Create and fill in the datatype */
  dteptr = (*newtype) = (MPI_Datatype) SBalloc( MPIR_dtes );
  dteptr->dte_type    = MPIR_INDEXED;
  dteptr->committed   = MPIR_FALSE;
  dteptr->basic       = MPIR_FALSE;
  dteptr->is_contig   = MPIR_FALSE;
  dteptr->ref_count   = 1;
  dteptr->align       = old_type->align;
  dteptr->old_type    = MPIR_Type_dup (old_type);
  dteptr->count       = count;
  dteptr->elements    = 0;

  /* Create indices and blocklens arrays and fill them */
  dteptr->indices     = ( int * ) MALLOC( count * sizeof( int ) );
  dteptr->blocklens   = ( int * ) MALLOC( count * sizeof( int ) );
  low                 = indices[0];
  high                = (indices[0]   * old_type->extent) + 
	                    (blocklens[0] * old_type->extent);
  for (i = 0; i < count; i++)  {
	dteptr->indices[i]    = indices[i] * old_type->extent;
	dteptr->blocklens[i]  = blocklens[i];
	ub = (dteptr->indices[i]   * old_type->extent) +
	     (dteptr->blocklens[i] * old_type->extent) ;
	if ( high < ub )
	  high = ub;
	low  = (low  > dteptr->indices[i]) ? dteptr->indices[i] : low;
	dteptr->elements     += blocklens[i];
  }

  /* Set the upper/lower bounds and the extent and size */
  dteptr->lb          = low * old_type->extent;
  dteptr->ub          = high;
  dteptr->extent      = dteptr->ub - dteptr->lb;
  dteptr->size        = dteptr->elements * old_type->size;
  dteptr->elements   *= old_type->elements;
  
  return (errno);
}

/*@
    MPI_Type_hindexed - Create an indexed datatype with offsets in bytes
@*/
int MPI_Type_hindexed( count, blocklens, indices, old_type, newtype )
int           count;
int           blocklens[];
MPI_Aint      indices[];
MPI_Datatype  old_type;
MPI_Datatype *newtype;
{
  MPIR_DATATYPE  *dteptr;
  MPI_Aint ub, high, low;
  int      i, errno = MPI_SUCCESS;

  /* Check for bad arguments */
  if ( ( (old_type == MPI_DATATYPE_NULL) && (errno = MPI_ERR_TYPE) )  ||
	   ( (old_type == (*newtype))        && (errno = MPI_ERR_TYPE) )  ||
	   ( (count    <= 0)                 && (errno = MPI_ERR_COUNT) ) )
	return MPIR_ERROR( MPI_COMM_NULL, errno,
					  "Error in MPI_TYPE_INDEXED" );
	
  /* Create and fill in the datatype */
  dteptr = (*newtype) = (MPI_Datatype) SBalloc( MPIR_dtes );
  dteptr->dte_type    = MPIR_HINDEXED;
  dteptr->committed   = MPIR_FALSE;
  dteptr->basic       = MPIR_FALSE;
  dteptr->is_contig   = MPIR_FALSE;
  dteptr->ref_count   = 1;
  dteptr->align       = old_type->align;
  dteptr->old_type    = MPIR_Type_dup (old_type);
  dteptr->count       = count;
  dteptr->elements    = 0;

  /* Create indices and blocklens arrays and fill them */
  dteptr->indices     = ( int * ) MALLOC( count * sizeof( int ) );
  dteptr->blocklens   = ( int * ) MALLOC( count * sizeof( int ) );
  low                 = indices[0];
  high                = indices[0] + (blocklens[0] * old_type->extent);
  for (i = 0; i < count; i++)  {
	dteptr->indices[i]    = indices[i];
	dteptr->blocklens[i]  = blocklens[i];
	ub = (dteptr->indices[i]) + (dteptr->blocklens[i] * old_type->extent) ;
	if ( high < ub )
	  high = ub;
	low  = (low  > dteptr->indices[i]) ? dteptr->indices[i] : low;
	dteptr->elements     += blocklens[i];
  }

  /* Set the upper/lower bounds and the extent and size */
  dteptr->lb          = low;
  dteptr->ub          = high;
  dteptr->extent      = high - low;
  dteptr->size        = dteptr->elements * old_type->size;
  dteptr->elements   *= old_type->elements;
  
  return (errno);
}

/*@
    MPI_Type_struct - Create a struct datatype
@*/
int MPI_Type_struct( count, blocklens, indices, old_types, newtype )
int           count;
int 	      blocklens[];
MPI_Aint      indices[];      
MPI_Datatype  old_types[];
MPI_Datatype *newtype;
{
  MPIR_DATATYPE  *dteptr;
  MPI_Aint ub, lb, high, low;
  int      i, errno = MPI_SUCCESS;

  /* Check for bad arguments */
  if ( (count    <= 0) && (errno = MPI_ERR_COUNT) )
	return MPIR_ERROR( MPI_COMM_NULL, errno,
					  "Error in MPI_TYPE_STRUCT" );
	
  /* Create and fill in the datatype */
  dteptr = (*newtype) = (MPI_Datatype) SBalloc( MPIR_dtes );
  dteptr->dte_type    = MPIR_STRUCT;
  dteptr->committed   = MPIR_FALSE;
  dteptr->basic       = MPIR_FALSE;
  dteptr->is_contig   = MPIR_FALSE;
  dteptr->ref_count   = 1;
  dteptr->align       = old_types[0]->align;
  dteptr->count       = count;
  dteptr->elements    = 0;
  dteptr->size        = 0;

  /* Create indices and blocklens arrays and fill them */
  dteptr->indices     = ( int * ) MALLOC( count * sizeof( int ) );
  dteptr->blocklens   = ( int * ) MALLOC( count * sizeof( int ) );
  dteptr->old_types   =
	( MPIR_DATATYPE ** ) MALLOC( count * sizeof( MPIR_DATATYPE *));
  low                 = indices[0] * old_types[0]->extent;
  high                = low + (blocklens[0] * old_types[0]->extent);
  for (i = 0; i < count; i++)  {
	dteptr->old_types[i]  = MPIR_Type_dup (old_types[i]);
	dteptr->indices[i]    = indices[i];
	dteptr->blocklens[i]  = blocklens[i];
	ub = (indices[i]   * old_types[i]->extent) + 
	     (blocklens[i] * old_types[i]->extent) ;
	lb = (indices[i] * old_types[i]->extent);
	if ( high < ub )
	  high = ub;
	if ( low  > lb )
	  low  = lb;
	dteptr->elements     += (blocklens[i] * old_types[i]->elements);
	dteptr->size         += (blocklens[i] * old_types[i]->extent);   
  }

  /* Set the upper/lower bounds and the extent and size */
  dteptr->lb          = low;
  dteptr->ub          = high;
  dteptr->extent      = high - low;
  
  return (errno);
}
